#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import re
import argparse
import cve_manager.defines as df
from collections      import defaultdict
from datetime         import datetime as dt
from ax.printer       import Printer
from ax.purefunctions import NearlyEqualShares
from cve_manager.conf import ConfParser

MATCH_VENDOR_PRODUCT = r'^(?P<vendor>[^:]+:)?(?P<product>[^\s]+)'
DISPUTED_REGEX = r"'^\\*\\* DISPUTED \\*\\*'"

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Get a defined value that corresponds to a given parameters

def GetDef(key, is_debug_mode):

	if key == 'CONF_DIR':
		return df.CONF_DIR_DEBUG if is_debug_mode else df.CONF_DIR

	if key == 'CONF_FILE':
		m = df.CVE_MANAGER
	elif key == 'CONF_FILE_MONITOR':
		m = df.CVE_MONITOR
	elif key == 'CONF_FILE_MAIL':
		m = df.CVE_MAIL
	else:
		return ''

	return df.CONF_FILES_DEBUG[m] if is_debug_mode else df.CONF_FILES[m]

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Form parsers for parameters that most of the modules are using

def NewArgParser(base=None, ptype='', description=''):

	if base:
		argparser = base
	else:
		argparser = argparse.ArgumentParser(description=description)

	if 'b' in ptype:
		argparser.add_argument(
			'-b', '--branches',
			metavar='BRANCH_NAME', type=str, nargs='+', default = [],
			help='Names of branches to be processed (all by default)'
			)

	if 'p' in ptype:
		argparser.add_argument(
			'-p', '--packages',
			metavar='PACKAGE_NAME', type=str, nargs='+', default = [],
			help='Names of packages to be processed (all by default)'
			)

	# 'm' for 'modes'
	if 'm' in ptype:
		# Second 'm' is for 'cve-monitor' that don't use some of the modes
		if ptype.count('m') == 1 :
			argparser.add_argument(
				'--plain',
				action='store_true',
				help='Use simpler formatting for an output'
				)
			argparser.add_argument(
				'--silent',
				action='store_true',
				help='Don\'t print anything to stdout'
				)
		argparser.add_argument(
			'--debug',
			action='store_true',
			help='Run in debug mode'
			)

	return argparser

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Prepare objects that every module needs right from beginning

# Get a Printer obj, a ConfParams obj and an err message (expected to be None)
def PureInit(args, monitor):

	for attr in ['debug'] +  ([] if monitor else ['plain', 'silent']):
		if not hasattr(args, attr):
			err = f'Attribute "{attr}" is missing'
			return None, None, err

	if monitor:
		printer = Printer()
		key = 'CONF_FILE_MONITOR'
	else:
		printer = Printer(plain=args.plain, silent=args.silent)
		key = 'CONF_FILE'

	conf, err = ConfParser(GetDef(key, args.debug)).Parse()
	if err:
		return printer, None, err

	return printer, conf, None

# The part that makes Init not pure
def Terminate(err, printer):
	if printer:
		printer.Err(err)
	else:
		print(f"[ERROR{(f': {err}' if err else '')}]")
	exit(1)

# Generate a Printer obj and a ConfParams obj and terminate if smth goes wrong
def Init(args, monitor=False):
	printer, conf, err = PureInit(args, monitor)
	if err:
		Terminate(err, printer)
	return printer, conf

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Split given cursor by a number of cores

def SplitData(cursor, n):

	data = []

	if n > 1:
		shares = NearlyEqualShares(cursor.rowcount, n)
		for share in shares:
			if share < 1:
				continue
			section = []
			for i, entry in enumerate(cursor):
				section.append(entry)
				if i == (share - 1):
					break
			data.append(section)
	else:
		data.append(cursor.fetchall())

	return data

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Revise a given CVE volumes

def ReviseCVEVolumes(vols, printer=None):

	# Forming a list of available CVE volumes (2002 - 202* + 'recent')
	cve_volumes = [str(i) for i in range(2002, dt.now().year + 1)]
	cve_volumes.append('recent')
	if not vols or 'all' in vols:
		return cve_volumes

	for vol in vols:
		if vol not in cve_volumes:
			if printer:
				printer.Warn(f'There is no "{vol}" CVE volume, choose from '
					f'"{cve_volumes[0]}"-"{cve_volumes[-2]}" and '
					f'"{cve_volumes[-1]}"')
			return []

	return vols

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Form a dict {tag: list_of_urls} out of a given string of tagged urls

def RefsDict(refs):

	refs_dict = defaultdict(list)

	for ref in refs.split():
		if not ref:
			continue
		m = re.match(r'^#([^#]*)#', ref)
		if m:
			tag = m.group(1)
			url = ref[len(m.group(0)):]
		else:
			tag = ''
			url = ref
		refs_dict[tag].append(url)

	return refs_dict

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Form a string of untagged urls out of a given string of tagged urls

def UntaggedRefs(refs, delim=' '):

	untagged_refs = ''
	refs_dict = RefsDict(refs)

	for _, urls in sorted(refs_dict.items()):
		untagged_refs += delim.join(urls) + delim

	return untagged_refs[:-len(delim)]

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Split given vul id into year and number (vul marker will be dropped)

def SplitVulId(vul_id):

	vul_id = vul_id[4:]

	if vul_id.count('-') == 1:
		return vul_id.split('-')

	return '', ''

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Form a "YYYY_MM_DD_hh_mm_ss" string from a [<year>, <month>, <day>, <hour>,
# <min>, <sec>] list

def DateTimeStr(list_date_time):

	list_date_time = [f'{el}' if i == 0 else f'{el:02d}'
		for i, el in enumerate(list_date_time)]

	return '_'.join(list_date_time)
