#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from os.path             import basename, expanduser
from configparser        import ConfigParser, ExtendedInterpolation
from cve_manager.defines import DATA_SOURCES

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Constants

DEFAULT_CONF_PATH = '/etc/cve-manager/cve-manager.conf'
COMMON_PARAMS_REQUIRED_FOR_ALL  = ('master_branch', 'download')
COMMON_PARAMS_REQUIRED_FOR_SOME = tuple()
COMMON_PARAMS_OPTIONAL = ('basealt', 'excluded_vulsrc', 'distro_lists_src',
	'distro_lists_bin', 'out_dir')
COMMON_PARAMS   = \
	COMMON_PARAMS_REQUIRED_FOR_ALL  + \
	COMMON_PARAMS_REQUIRED_FOR_SOME + \
	COMMON_PARAMS_OPTIONAL
COMMON_PARAMS_MULTI = ('excluded_vulsrc',)
COMMON_PARAMS_PATHS = ('basealt', 'download', 'out_dir')
DB_CON_PARAMS   = ('host', 'user', 'password', 'database')
MAIL_CON_PARAMS = ('host', 'port', 'user', 'password', 'raddress')
DB_CON_SEC = 'db_connection'
COMMON_SEC = 'common'
MAIL_SEC   = 'mail'
BASIC_SECTIONS = (DB_CON_SEC, COMMON_SEC, MAIL_SEC)
SMTP_CON_SEC   = 'smtp_connection'
RECIPIENTS_SEC = 'recipients'
LOCAL_SYS = 'local_sys'

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Basic parser of ini-files, that more prolific classes inherit from

class IniParser():

	def __init__(self, conf_path=DEFAULT_CONF_PATH):

		self.config = None
		self.conf_path = conf_path

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Convert "'a', 'b', ..." (or '"a", "b", ...') str into [a, b, ...] list

	def SplitValue(self, complex_value):

		return [v.strip('\'" \t\n') for v in complex_value.split(',')]

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check that a section with a given name has all required items
	# and return that section

	def GetSection(self, secname, required_items=[]):

		if not self.config.has_section(secname):
			err = f'There is no {secname} section'
			return None, err

		sec = self.config[secname]

		# Checking that all required params are there
		for item in required_items:
			if not sec.get(item):
				err = f'Can\'t get value of "{secname}:{item}" item ' \
					f'from "{self.conf_path}"'
				return None, err

		return sec, None

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Read a conf file and return an error message

	def __Read(self):

		# Each conf file can be read only once
		if self.config:
			return f'Conf file {self.conf_path} has been parsed already'

		# Trying to read the configuration file
		try:
			self.config = ConfigParser(interpolation=ExtendedInterpolation())
			with open(self.conf_path, 'r'):
				pass
		except PermissionError:
			return 'You don\'t have the rights to read the configuration ' \
				f'file, make sure that "{self.conf_path}" group is "cve" ' \
				'and you are a member of that group'
		except FileNotFoundError:
			return f'Can\'t find "{self.conf_path}" configuration file'

		# Parsing the conf
		self.config.read(self.conf_path)
		if not self.config:
			return 'The configuration file seems to be empty'

		return None

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Parse a conf file using a given dict of functions and return
	# a ConfParams obj and an err message

	def Parse(self, functions):

		err = self.__Read()
		if err:
			return None, err

		params = {}

		for sec, f in functions.items():
			items, err = f()
			if err:
				return None, err
			params[sec] = items

		return ConfParams(params), None

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Parser for files used to configure all of the modules

class ConfParser(IniParser):

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get MySQL params from a parsed conf file

	def __FormMySQLConfig(self):

		sec, err = super().GetSection(DB_CON_SEC, DB_CON_PARAMS)
		if err:
			return {}, err

		params = {}

		for k, v in sec.items():
			v = v.strip('\'"')
			if k == 'host':
				# MySQL connector doesn't need a port number
				params[k] = v.split(':')[0]
			elif k in ('user', 'password', 'database'):
				params[k] = v

		return params, None

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get various common params from a parsed conf file

	def __FormCommon(self):

		req = COMMON_PARAMS_REQUIRED_FOR_ALL

		# cve-monitor may not require some of the common params
		# that other modules do require
		if 'monitor' not in basename(self.conf_path):
			req += COMMON_PARAMS_REQUIRED_FOR_SOME

		sec, err = super().GetSection(COMMON_SEC, req)
		if err:
			return {}, err

		common = {}

		for k, v in sec.items():
			if k not in COMMON_PARAMS:
				continue
			if k in COMMON_PARAMS_MULTI:
				common[k] = [el.strip('\'"') for el in v.split(',')]
			if k in COMMON_PARAMS_PATHS:
				common[k] = expanduser(v)
			else:
				common[k] = v.strip('\'"')

		return common, None

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get available branches from a parsed conf file

	def __FormAvailableBranches(self):

		branches = []

		for sec in self.config.sections():
			# Skipping basic sections and repetitions
			if sec in BASIC_SECTIONS or any(b['name'] == sec for b in branches):
				continue
			values = {}
			for k, v in self.config[sec].items():
				if k == 'seniors':
					values[k] = super().SplitValue(v)
				else:
					values[k] = v
			branches.append({
				'name'    : sec,
				'branch'  : values.get('branch'),
				'seniors' : values.get('seniors', []),
				})

		if not branches:
			err = \
				f'There isn\'t a single branch description in {self.conf_path}'
			return [], err

		return branches, None

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Parse a conf file using special dict of functions
	# <section name> : <function to handle items of that section>

	def Parse(self):

		functions = {
			DB_CON_SEC : self.__FormMySQLConfig,
			COMMON_SEC : self.__FormCommon,
			'branches' : self.__FormAvailableBranches
			}

		return super().Parse(functions)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Parser for a file used to configure the mailer

class MailConfParser(IniParser):

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get connection configuration and categories of recipients

	def __FormMailConnection(self):

		sec, err = super().GetSection(SMTP_CON_SEC, MAIL_CON_PARAMS)
		if err:
			return {}, err

		return sec, None

	def __FormMailRecipients(self):

		sec, err = super().GetSection(RECIPIENTS_SEC)
		if err:
			return {}, err

		recipients = {}

		for category, addresses in sec.items():
			recipients[category] = super().SplitValue(addresses)
		if not recipients:
			err = f'There is no "{RECIPIENTS_SEC}" section'
			return {}, err

		return recipients, None

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Parse a conf file using special dict of functions
	# <section name> : <function to handle items of that section>

	def Parse(self):

		functions = {
			SMTP_CON_SEC   : self.__FormMailConnection,
			RECIPIENTS_SEC : self.__FormMailRecipients
			}

		return super().Parse(functions)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Class for storing and accessing params of previously parsed config file

class ConfParams():

	def __init__(self, params):

		self.params = params

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check availability of given branches and return err if some not available

	def __CheckBranch(self, branch):

		for available_branch in self.params.get('branches'):
			if branch == available_branch['name']:
				return None

		return f'There is no "{branch}" branch'

	def __CheckBranches(self, branches):

		for branch in branches:
			err = self.__CheckBranch(branch)
			if err:
				return err

		return None

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a list of values of given sections

	def Get(self, sections):

		res = []

		for sec in sections:
			sec = self.params.get(sec)
			if sec:
				res.append(sec)

		return res

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a dict of group names of given branches

	def GetSeniors(self, branches):

		seniors = {}

		for conf_branch in self.params['branches']:
			for given_branch_name in branches:
				if conf_branch['name'] == given_branch_name:
					seniors[given_branch_name] = conf_branch['seniors']
					break

		return seniors

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get a name of a repository used by the the local system

	def GetLocalSysBranch(self):

		for b in self.params['branches']:
			if b.get('name') == LOCAL_SYS:
				branch_name = b.get('branch')
				if branch_name:
					return branch_name
				break

		return self.params[COMMON_SEC]['master_branch']

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Try to select given branches out of all available branches, or
	# return all available branches if given list is empty

	def SelectBranches(self, branches=[]):

		selection = []

		if branches:
			err = self.__CheckBranches(branches)
			if err:
				return [], err
			for b in branches:
				if b not in selection:
					selection.append(b)
		else:
			for b in self.params['branches']:
				if b.get('name') not in selection:
					selection.append(b.get('name'))

		return selection, None

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Try to select given data sources out of all available ones, or
	# return all available data sources if given list is empty

	def SelectDataSources(self, data_sources=[]):

		if not data_sources:
			data_sources = DATA_SOURCES
		else:
			for s in data_sources:
				if s not in DATA_SOURCES:
					err = f'There is no {s} data source'
					return [], err

		excluded = self.params.get(COMMON_SEC, {}).get('excluded_vulsrc', [])
		selection = [s for s in data_sources if s not in excluded]

		return selection, None
