#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import mysql.connector   as mysql
from cve_manager.common  import DISPUTED_REGEX
from cve_manager.defines import DATA_SOURCES, VUL_MARKERS

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Form a prefix for a history table that is of a given type, of a given type of
# issues and of a branch with a given name

def PrefixOfHistoryTables(type_of_history, type_of_issues, branch):

	if type_of_history == 'issues':
		return '_'.join([branch, type_of_issues, 'issues', 'unfixed', ''])
	elif type_of_history == 'map':
		return '_'.join([branch, 'map', 'history', ''])

	return ''

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Get a string of placeholders for a query

def Placeholders(n):

	return ", ".join(('%s' for _ in range(n)))

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Form a query to fetch issues of a given branch & type

def QueryToFetchIssues(branch, type_of_issues, condition='', acl=True, fix=True,
	scores=True, disputed=False, concat_year_num=False):

	vul_import_table = f'{type_of_issues}_vul_import'
	issues_table = '_'.join([branch, type_of_issues, 'issues'])
	src_table = branch + '_src'

	return ('SELECT ' +
		(f'CONCAT(\'{VUL_MARKERS.get(type_of_issues, "")}\', '
				f'{issues_table}.vul_year, "-", {issues_table}.vul_num) vul_id, '
			if concat_year_num else
			f'{issues_table}.vul_year, {issues_table}.vul_num, ') +
		f'{src_table}.name, '
		f'{src_table}.{type_of_issues}_name, '
		f'{src_table}.ver, '
		f'{src_table}.rel, '
		f'{issues_table}.vul_ver '
		f'{f", {vul_import_table}.cvss_score_v2 " if scores else ""}'
		f'{f", {vul_import_table}.cvss_score_v3 " if scores else ""}'
		f'{f", REGEXP_LIKE({vul_import_table}.summary, {DISPUTED_REGEX})" if disputed else ""}'
		f'{f", {issues_table}.fix " if fix else ""}'
		f'{f", {src_table}.acl " if acl else ""}'
		f'FROM {issues_table} INNER JOIN {src_table} '
			f'ON {issues_table}.package_id = {src_table}.id '
		f'INNER JOIN {vul_import_table} '
			f'ON {issues_table}.vul_year = {vul_import_table}.vul_year '
			f'AND {issues_table}.vul_num = {vul_import_table}.vul_num '
		f'{f"WHERE {condition}" if condition else ""}')

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Interface for MySQL CVE database

class DB():

	def __init__(self, printer=None, silent=False):

		self.con = None
		self.cursor = None
		self.buf_cursors = []
		self.user = None
		self.password = None
		self.database = None
		self.printer = printer
		self.silent = silent

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Close MySQL connection

	def __exit__(self):

		self.Disconnect()

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Print using a Printer obj or the regular way

	def __PrintErr(self, msg):

		if self.silent:
			return
		if self.printer:
			self.printer.Err(msg)
		else:
			print(f'[ERROR: {msg}]')

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Close prev cursor instance and create a new one

	def __RecreateCursor(self, n=-1):

		if n >= 0:
			if len(self.buf_cursors) > n:
				self.buf_cursors[n].close()
				self.buf_cursors[n] = self.con.cursor(buffered=True)
			else:
				while len(self.buf_cursors) <= n:
					self.buf_cursors.append(self.con.cursor(buffered=True))
		else:
			if self.cursor:
				self.cursor.close()
				self.cursor = None
			if self.con:
				self.cursor = self.con.cursor()

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check DB existence

	def __DBExists(self):

		self.Cursor("SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA "
			f"WHERE SCHEMA_NAME = '{self.database}'")

		for res in self.cursor:
			return True

		return False

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Recreate DB

	def __RecreateDB(self):

		for op in ("DROP", "CREATE"):
			try:
				if op == 'DROP' and not self.__DBExists():
					continue
				self.Cursor(f"{op} DATABASE {self.database}")
			except mysql.Error as err:
				self.__PrintErr(str(err))
				return False

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Connect to a CVE database

	def Connect(self, mysql_config, n_bcurs=0, recreate_db=False):

		if not self.Disconnect():
			return False

		self.user     = mysql_config['user']
		self.password = mysql_config['password']
		self.database = mysql_config['database']
		mysql_config['auth_plugin'] = 'mysql_native_password'

		try:
			if recreate_db:
				del mysql_config['database']
				self.con = mysql.connect(**mysql_config, buffered=True)
				if not self.__RecreateDB() or not self.Disconnect():
					return False
				mysql_config['database'] = self.database
			# For some reason passing buffered flag per cursor doesn't work
			self.con = mysql.connect(**mysql_config, buffered=True)
		except mysql.Error as err:
			if err.errno == mysql.errorcode.ER_BAD_DB_ERROR:
				self.__PrintErr(f'Database "{self.database}" doesn\'t exist')
			elif err.errno == mysql.errorcode.ER_ACCESS_DENIED_ERROR:
				self.__PrintErr('Wrong database/user/password')
			else:
				self.__PrintErr(str(err))
			return False

		self.__RecreateCursor()
		for i in range(0, n_bcurs):
			self.__RecreateCursor(i)

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Disconnect from a CVE database

	def Disconnect(self):

		if self.con:
			try:
				# Closing unbufferd cursor
				if self.cursor:
					self.cursor.close()
					self.cursor = None
				# Closing all bufferd cursors
				if self.buf_cursors:
					for c in self.buf_cursors:
						if c:
							c.close()
					self.buf_cursors = []
				# Closing connection
				if self.con and self.con.is_connected():
					self.con.close()
					self.con = None
			except mysql.Error as err:
				self.__PrintErr(str(err))
				return False

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Use unbuffered or one of the buffered cursors

	def Cursor(self, query, tables=tuple(), data=[], n=-1):

		for table in tables:
			table_exists = self.TableExists(table)
			if not table_exists:
				if table_exists != None:
					self.__PrintErr(f'Table "{table}" doesn\'t exist')
				return None

		self.__RecreateCursor(n)

		try:
			if n >= 0:
				if n > len(self.buf_cursors):
					self.__PrintErr(f'There is no cursor with index {n}')
					return None
				if data:
					self.buf_cursors[n].executemany(query, data)
					self.con.commit()
				else:
					self.buf_cursors[n].execute(query)
				return self.buf_cursors[n] if self.buf_cursors[n] else []
			if data:
				self.cursor.executemany(query, data)
				self.con.commit()
			else:
				self.cursor.execute(query)

		except mysql.Error as err:
			self.__PrintErr(str(err))
			return None

		return self.cursor if self.cursor else []

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Select specified columns from a specified table

	def Select(self, cols, table, cond=None, data=[], n=-1):

		query = f"SELECT {', '.join(cols)} FROM {table} " + \
			(f"WHERE {cond}" if cond else "")

		return self.Cursor(query, tables=(table,), data=data, n=n)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get list of tables that satisfy given condition

	def ShowTables(self, condition=''):

		query  = f'SHOW TABLES {f"WHERE {condition}" if condition else ""}'
		cursor = self.Cursor(query)
		tables = []

		for table_name, in cursor:
			# Converting from byte arrays if needed
			tables.append(table_name if type(table_name) is str
				else table_name.decode('utf-8'))

		return tables

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get list of 'history' tables (in byte arrays)

	def ShowHistoryTables(self, type_of_history, type_of_issues, branch):

		prefix = PrefixOfHistoryTables(type_of_history, type_of_issues, branch)
		condition = f"Tables_in_{self.database} REGEXP '{prefix}'"

		return self.ShowTables(condition)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Check given table existence in DB

	def TableExists(self, table):

		if self.Cursor(f"SHOW TABLES LIKE '{table}'") == None:
			return None

		for res in self.cursor:
			return True

		return False

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Drop given table if it exists in database

	def DropTable(self, table):

		table_exists = self.TableExists(table)
		if table_exists == None:
			return False

		if table_exists and self.Cursor(f'DROP TABLE {table}') == None:
			return False

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Recreate table after checking that it exists and dropping it if it exists

	def RecreateTable(self, table, columns):

		table_exists = self.TableExists(table)
		if table_exists == None:
			return False

		if table_exists and not self.DropTable(table):
			return False

		query = f"CREATE TABLE {table} ({', '.join(columns)})"

		return self.Cursor(query) != None

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Count number of table's entries

	def CountRows(self, table, target='*', cond=''):

		if self.cursor:
			cursor = self.Select(f'COUNT({target})', table, cond)
			if cursor and cursor.rowcount > 0:
				res = cursor.fetchone()
				return res[0]

		return -1

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get all packages of a given branch

	def QueryPackages(self, branch, src_packages={}, bin_packages={}, target='',
			custom_condition='', mapping_name=False, url=False, fixes=False,
			n=-1):

		table = f'{branch}_src'
		cols = ['id', 'name', 'ver', 'rel']

		if mapping_name:
			cols += [f'{t}_name'
				for t in ((target,) if target else DATA_SOURCES)]
		if url:
			cols.append('url')
		if fixes:
			cols.append(f'{VUL_MARKERS.get(target, "")[:-1].lower()}_fixes')

		cond = ''

		# Add given condition
		def AddCondition(c):
			nonlocal cond
			cond += f'AND ({c}) ' if cond else f'({c})'
		# - nested func

		# Form a condition to select packages that have specified names
		def PackagesCondition(packages):
			c = ''
			for package in packages:
				if package:
					c += f"{'OR ' if c else ''}name = '{package}' "
			return c
		# - nested func

		if src_packages:
			AddCondition(PackagesCondition(src_packages))

		if bin_packages:
			AddCondition(f'id in (SELECT src_id FROM {branch}_bin WHERE ' +
				f'{PackagesCondition(bin_packages)})')

		if custom_condition:
			AddCondition(custom_condition)

		return self.Select(cols, table, cond, n=n)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Get issues of a given branch & type (arglist matches an arglist
	# of the QueryToFetchIssues function)

	def QueryIssues(self, *args, **kwargs):

		return self.Cursor(QueryToFetchIssues(*args, **kwargs))
