#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from re                            import match
from collections                   import defaultdict
from cve_manager.common            import MATCH_VENDOR_PRODUCT
from cve_manager.defines           import NVD_DATA_SRC
from cve_manager.knowledge_handler import KnowledgeHandler
from cve_manager.url               import ParseAndJoinURL

MATCH_PACKAGE_NAME_AND_URLS = r'(?P<package>[^\s]+)(\s+\((?P<urls>[^\)]*)\))?'

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

class IgnoredPairs(KnowledgeHandler):

	def __init__(self, conf_dir, data_source):

		file_name = f'{"cpe" if data_source == NVD_DATA_SRC else data_source}-mapping-ignore.csv'

		super().__init__(conf_dir, file_name)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
	# Parse a file that contain mapping pairs that should be ignored

	def Read(self):

		ok, msg = super().Check()
		if not ok:
			return None, msg

		ignore = defaultdict(lambda: defaultdict(lambda: defaultdict(set)))
		first_line = True

		with open(self.file_path, 'r') as f:
			while True:
				line = f.readline()
				if not line:
					break
				if first_line:
					# Omitting the header
					first_line = False
					continue
				values = [el.strip() for el in line.split(', ')]
				# Branch names can be specified in the third column
				if len(values) == 2:
					branches = {''}
				elif len(values) == 3:
					branches = set(values[2].split())
				else:
					continue
				# Extacting a package name and a package url
				m = match(fr'^{MATCH_PACKAGE_NAME_AND_URLS}$', values[0])
				if not m:
					continue
				package_name = m.group('package')
				package_urls = \
					[ParseAndJoinURL(url) for url in m.group('urls').split()] \
					if m.group('urls') else ['']
				# For each CPE separated by spaces
				for cpe in values[1].split():
					# Extracting a product vendor and a product name
					m = match(fr'^{MATCH_VENDOR_PRODUCT}$', cpe)
					if not m:
						continue
					vendor = m.group('vendor')[:-1] if m.group('vendor') else ''
					product = m.group('product')
					# For each branch
					for branch in branches:
						for package_url in package_urls:
							ignore[branch][package_name][package_url] \
								.add((vendor, product))

		if ignore:
			count = sum(len(s) for d1 in ignore.values() for d2 in d1.values()
				for s in d2.values())
			msg = f'Ignoring {count} pair{"s" if count != 1 else ""}'
		else:
			msg = 'Not ignoring a thing'

		return ignore, msg
