#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import re
from urllib.parse            import urlparse
from cve_manager.url_github  import Fmt as FmtGitHub
from cve_manager.url_sforge  import Fmt as FmtSourceForge
from cve_manager.url_gnu     import Fmt as FmtGNU
from cve_manager.url_apache  import Fmt as FmtApache
from cve_manager.url_debian  import Fmt as FmtDebian
from cve_manager.url_gnome   import Fmt as FmtGnome
from cve_manager.url_fedora  import Fmt as FmtFedoraHosted
from cve_manager.url_freedt  import Fmt as FmtFreeDesktop
from cve_manager.url_mozilla import Fmt as FmtMozilla
from cve_manager.url_other   import FmtTwoPart, FmtPython, FmtGoogle

# Reg expressions and functions to treat URLs that satisfy those expressions
SPECIAL_CASES = [
	(r'(^|\.)github\.(com|io)$',   FmtGitHub),
	(r'(^|\.)(sf|sourceforge)\.net$', FmtSourceForge),
	(r'(^|\.)pypi\.python\.org$',  FmtPython),
	(r'(^|\.)gnu\.org$',           FmtGNU),
	(r'(^|\.)apache\.org$',        FmtApache),
	(r'(^|\.)debian\.org$',        FmtDebian),
	(r'(^|\.)gnome\.org$',         FmtGnome),
	(r'(^|\.)fedorahosted\.org$',  FmtFedoraHosted),
	(r'(^|\.)(bitbucket\.org|gitlab\.com)$', FmtTwoPart),
	(r'(^|\.)freedesktop\.org$',   FmtFreeDesktop),
	(r'(^|\.)mozilla\.org$',       FmtMozilla),
	(r'(^|\.)code\.google\.com$',  FmtGoogle),
	]

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Get a dict containing of netloc, path and query of a given URL-address

def ParseURL(url):

	# Handling a situation when a scheme is not specified
	if not re.search(r'^[A-Za-z]+://', url):
		url = 'x://' + url

	parsed_url = urlparse(url)
	netloc = re.sub(r'^www\.', '', parsed_url.netloc.lower())
	path   = re.sub(
		r'/downloads?(/|$)',      '/', re.sub(
		r'\.(s?html?|php|git)$',  '',  re.sub(
		r'/$',                    '',  re.sub(
		r'/index\.(s?html?|php)', '',  parsed_url.path.lower())))
		)
	query = re.sub(r';?a=(summary|tree|tag)$', '', parsed_url.query.lower())

	for regex, func in SPECIAL_CASES:
		if re.search(regex, netloc):
			netloc, path, query = func(netloc, path, query)
			break

	return {
		'netloc' : netloc,
		'path'   : path,
		'query'  : query,
		}

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Join a given list of components of a parsed URL

def JoinURL(components):

	url = ''

	for el in components:
		el = ''.join(el.split())
		if not el:
			continue
		url += el if el[0] != '/' else el[1:]
		if url[-1] != '/':
			url += '/'

	return url[:-1]

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Parse a specified URL and join it's components together

def ParseAndJoinURL(url):

	parsed_url = ParseURL(url)

	return JoinURL([parsed_url['netloc'], parsed_url['path']])
