#!/usr/bin/python3

#	cve-manager : CVE management tool
#	Copyright (C) 2017-2022 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import re

BASIC_NETLOC = 'apache.org'
BASIC_NETLOC_LEN = len(BASIC_NETLOC)

# There is a lot of useless stuff in Apache URLs
USELESS_PARTS = [
	r'(v|v?[0-9\._\-]+)((\-|_)[^/]+)?(/.*|$)',
	r'(docs|news|changes((\-|_)[^/]+)?|security((\-|_)[^/]+)?|blog|announcements)(/.*|$)',
	r'dist/',
	r'([^/]+(\-|_))?release(s|(\-|_)[^/]+)?(/.*|$)',
	r'([^/]+(\-|_))?downloads?((\-|_)[^/]+)?(/|$)',
	r'repos(/|$)',
	r'(binaries|documentation|changelog|dev|mirror|en(\-|_)us)(/|$)',
	r'[^/]+\.(txt|cgi|jspa?)$',
	]

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Format an Apache URL

def Fmt(netloc, path, query):

	prefix = netloc[:-(BASIC_NETLOC_LEN + 1)] \
		if len(netloc) > BASIC_NETLOC_LEN else ''
	product = ''

	# Apache URL can be in one of the following formats:
	# a) svn.apache.org/(viewvc|repos)/(apr|asf|db)/<product>(/.*|$)
	if prefix == 'svn':
		splitted_path = path.split('/')
		if len(splitted_path) > 3 and \
				splitted_path[1] in ('viewvc', 'repos') and \
				splitted_path[2] in ('asf', 'apr', 'db'):
			product = splitted_path[3]
			path = ''
	# b) lists.apache.org/thread.html/<hash>(@|%40)%3(<keyword>.)?<product>.apache.org%3e
	elif prefix == 'lists':
		m = re.match(r'.+(@|%40)%3([^\.]+\.)?([^\.]+)\.apache\.org', path)
		if m:
			product = m.group(3)
			path = ''
	# c) issues.apache.org/jira/browse/<product>(-[0-9]+)?(/.*|$)
	elif prefix == 'issues':
		m = re.match(r'^/jira/browse/([^/]+)', path)
		if m:
			product = re.sub(r'\-[0-9]+$', '', m.group(1))
			path = ''
	# d) repository.apache.org/content/repositories/releases/org/apache/<product>(/.*|$)
	elif prefix == 'repository':
		m = re.match(r'^/content/repositories/releases/org/[^/]+/([^/]+)', path)
		if m:
			product = m.group(1)
			path = ''
	# e) cwiki.apache.org/confluence/display/<product>(/.*|$)
	elif prefix == 'cwiki':
		m = re.match(r'^/[^/]+/display/([^/]+)', path)
		if m:
			product = m.group(1)
			path = ''
	# f) blogs.apache.org/<product>/.*
	elif prefix == 'blogs':
		m = re.match(r'^/([^/]+)', path)
		if m:
			product = m.group(1)
			path = ''
	# g) git-wip-us.apache.org/(...)?p=impala.git
	elif prefix == 'git-wip-us':
		if query:
			m = re.match(r'p=([^\.&]+)', query)
			if m:
				product = m.group(1)
				path = ''
	else:
		# h) (archive|dist).apache.org/(...)/<product>(/.*|$)
		if prefix in ('dist', 'archive'):
			netloc = BASIC_NETLOC
		else:
			product = prefix
		# i) (<product>.)?apache.org(/.*|$)
		# j) apache.org/(...)/<product>(/.*|$)
		if path:
			# Cutting useless combinations from a path
			for useless in USELESS_PARTS:
				path = re.sub(r'/' + useless, '/', path)
				if not path:
					break
			path = re.sub(r'/$', '', path)
			# Converting 'apache.org/<product>' to '<product>.apache.org'
			if not product and path.count('/') == 1:
				product = path[1:]
				path = ''

	if product:
		return f'{product}.{BASIC_NETLOC}', path, ''

	return netloc, path, query
